/*
 * File: Shape.cs
 *
 * Authors: Aaron Wolin, Devin Smith, Jason Fennell, and Max Pflueger.
 * Harvey Mudd College, Claremont, CA 91711.
 * Sketchers 2006.
 * 
 * Use at your own risk.  This code is not maintained and not guaranteed to work.
 * We take no responsibility for any harm this code may cause.
 */

using System;
using System.Collections.Generic;


namespace Sketch
{
	/// <summary>
	/// Shape class.
	/// </summary>
	[Serializable]
	public class Shape : IComparable
	{
		#region INTERNALS

		/// <summary>
		/// Substrokes
		/// </summary>
		private List<Substroke> substrokes;

		/// <summary>
		/// Shapes
		/// </summary>
		private List<Shape> shapes;

		/// <summary>
		/// Parent shape
		/// </summary>
		private Shape parentShape;

		/// <summary>
		/// Xml attributes of the Shape
		/// </summary>
		public XmlStructs.XmlShapeAttrs XmlAttrs;

		/// <summary>
		/// This is a variable added by Eric
		/// </summary>
		private int index;

        /// <summary>
        /// This is the name of the current shape 
        /// <example>A</example>
        /// <example>Wire0</example>
        /// <example>And1</example>
        /// </summary>
        private string name;

        /// <summary>
        /// Lists of shapes connected to this shape
        /// </summary>
        private List<Shape>[] directionallyConnectedShapes;
        private Dictionary<Shape, string> connectedShapes;

        /// <summary>
        /// Holds the possible orientations of the symbol.
        /// </summary>
        public enum Orientation
        {
            Unknown = -1,
            Up,
            Down,
            Left,
            Right,
        }

        // constants for indexint into directionallyconnectedshapes
        private int UP = 0;
        private int DOWN = 1;
        private int LEFT = 2;
        private int RIGHT = 3;

        private Orientation orient;

		#endregion

		#region CONSTRUCTORS
		
		/// <summary>
		/// Constructor
		/// </summary>
		public Shape() :
			this(new List<Substroke>(), new XmlStructs.XmlShapeAttrs(true))
		{
			//Calls the main constructor
		}

		/// <summary>
		/// Constructor. Create from Shape. Shallow copy of subshapes, deep copy of substrokes.
		/// </summary>
		/// <param name="shape">A Shape.</param>
		public Shape(Shape shape)
		{
            this.name = shape.name;
			index = -1;
			index = shape.index;
			List<Substroke> newSubstrokes = new List<Substroke>(shape.substrokes.Count);
			List<Shape> newShapes = new List<Shape>(shape.shapes.Count);
			foreach (Substroke s in shape.substrokes)
			{
				Substroke newS = s.CloneConstruct();
				List<Shape> parents = new List<Shape>(s.ParentShapes);
				// Switch the original shape with the clone in the parents list
				parents.Remove(shape);
				parents.Add(this);
				newS.ParentShapes = parents;
				newSubstrokes.Add(newS);

			}

            this.connectedShapes = shape.connectedShapes;
            this.directionallyConnectedShapes = shape.directionallyConnectedShapes;
            this.orient = shape.orient;

			shapes = new List<Shape>(shape.shapes);
			substrokes = new List<Substroke>(shape.substrokes.Count);
			foreach (Substroke s in newSubstrokes)
				AddSubstroke(s);
			XmlAttrs = shape.XmlAttrs.Clone();
		}


		/// <summary>
		/// Construct a Shape with given Shapes, Substrokes, and XML attributes.
		/// </summary>
		/// <param name="shapes">Shapes to add</param>
		/// <param name="substrokes">Substrokes to add</param>
		/// <param name="XmlAttrs">XML attributes of the Shape</param>
		public Shape(Shape[] shapes, Substroke[] substrokes, XmlStructs.XmlShapeAttrs XmlAttrs)
			: this(new List<Shape>(shapes), new List<Substroke>(substrokes), XmlAttrs)
		{
			// Calls the main constructor
		}

	
		/// <summary>
		/// Construct a Shape with given Shapes, Substrokes, and XML attributes.
		/// </summary>
		/// <param name="shapes">Shapes to add</param>
		/// <param name="substrokes">Substrokes to add</param>
		/// <param name="XmlAttrs">XML attributes of the Shape</param>
		public Shape(List<Shape> shapes, List<Substroke> substrokes, XmlStructs.XmlShapeAttrs XmlAttrs)
		{
			this.shapes = new List<Shape>();
			this.substrokes = new List<Substroke>();
			this.parentShape = null;

			this.XmlAttrs = XmlAttrs;

            this.AddSubstrokes(substrokes);
			
            this.AddShapes(shapes);

            this.connectedShapes = new Dictionary<Shape, string>();
            this.directionallyConnectedShapes = new List<Shape>[4];
            this.orient = Orientation.Unknown;
		}

		/// <summary>
		/// Creates a new Shape with given substrokes, XmlAttrs
		/// </summary>
		/// <param name="substrokes">Substrokes to add</param>
		/// <param name="XmlAttrs">XML Attributes to add</param>
		public Shape(List<Substroke> substrokes, XmlStructs.XmlShapeAttrs XmlAttrs)
			: this(new List<Shape>(), substrokes, XmlAttrs)
		{
			// Nothing to do here
		}

		#endregion

		#region ADD TO SHAPE

		#region ADD SUBSTROKE(S)

		/// <summary>
		/// Add a Substroke to this shape
		/// </summary>
		/// <param name="substroke">A Substroke</param>
		public void AddSubstroke(Substroke substroke)
		{
			//Console.WriteLine(this.XmlAttrs.Id + ": ADD :" + substroke.XmlAttrs.Id);
            if (!this.substrokes.Contains(substroke))
            {
                int low = 0;
                int high = this.substrokes.Count - 1;
                int mid;
                while (low <= high)
                {
                    mid = (high - low) / 2 + low;
                    if (substroke.XmlAttrs.Time.Value < this.substrokes[mid].XmlAttrs.Time.Value)
                        high = mid - 1;
                    else
                        low = mid + 1;
                }

                this.substrokes.Insert(low, substroke);

                if (substroke.ParentShapes != null && !substroke.ParentShapes.Contains(this))
                    substroke.ParentShapes.Add(this);

                UpdateAttributes();
            }
            else
            {
				Console.WriteLine("Warning: cannot add substroke " + substroke.XmlAttrs.Id);
				Console.WriteLine("to shape " + XmlAttrs.Id + " because it's already there!");
                //throw new Exception("substroke already exists in shape");
            }
		}

        /// <summary>
        /// Add a Substroke to this shape
        /// </summary>
        /// <param name="substroke">A Substroke</param>
        public void AddSubstrokeByID(Substroke substroke)
        {
            //Console.WriteLine(this.XmlAttrs.Id + ": ADD :" + substroke.XmlAttrs.Id);
            foreach(Substroke substroke1 in substrokes)
                if(substroke.Id == substroke1.Id)
                {
                    Console.WriteLine("Warning: cannot add substroke " + substroke.XmlAttrs.Id);
                    Console.WriteLine("to shape " + XmlAttrs.Id + " because it's already there!");
                    //throw new Exception("substroke already exists in shape");
                    return;
                }

            int low = 0;
            int high = this.substrokes.Count - 1;
            int mid;
            while (low <= high)
            {
                mid = (high - low) / 2 + low;
                if (substroke.XmlAttrs.Time.Value < this.substrokes[mid].XmlAttrs.Time.Value)
                    high = mid - 1;
                else
                    low = mid + 1;
            }

            this.substrokes.Insert(low, substroke);

            if (substroke.ParentShapes != null && !substroke.ParentShapes.Contains(this))
                substroke.ParentShapes.Add(this);

            UpdateAttributes();

        }

		
		/// <summary>
		/// Add Substrokes to this shape
		/// </summary>
		/// <param name="substrokes">The Substrokes</param>
		public void AddSubstrokes(Substroke[] substrokes)
		{
			int length = substrokes.Length;
			for (int i = 0; i < length; ++i)
			{
				this.AddSubstroke(substrokes[i]);
			}
		}

        /// <summary>
        /// Add Substrokes to this shape by ID
        /// </summary>
        /// <param name="substrokes">The Substrokes</param>
        public void AddSubstrokesByID(Substroke[] substrokes)
        {
            int length = substrokes.Length;
            for (int i = 0; i < length; ++i)
            {
                this.AddSubstrokeByID(substrokes[i]);
            }
        }
		
		/// <summary>
		/// Add Substrokes to this shape
		/// </summary>
		/// <param name="substrokes">The Substrokes</param>
		public void AddSubstrokes(List<Substroke> substrokes)
		{
			int length = substrokes.Count;
			for (int i = 0; i < length; ++i)
				this.AddSubstroke(substrokes[i]);
		}

		#endregion

		#region ADD SHAPE(S)

		/// <summary>
		/// Add a Shape to this Shape
		/// </summary>
		/// <param name="shape">The Shape to add</param>
		public void AddShape(Shape shape)
		{
			if (!this.shapes.Contains(shape))
			{
				int low = 0;
				int high = this.shapes.Count - 1;
				int mid;
				while (low <= high)
				{
					mid = (high - low) / 2 + low;
					if (shape.XmlAttrs.Time.Value < this.shapes[mid].XmlAttrs.Time.Value)
						high = mid - 1;
					else
						low = mid + 1;
				}
		
				this.shapes.Insert(low, shape);
				shape.ChangeParentShape(this);

				UpdateAttributes();
			}
		}


		/// <summary>
		/// Add Shapes to this Shape
		/// </summary>
		/// <param name="shapes">The Shapes</param>
		public void AddShapes(Shape[] shapes)
		{
			int length = shapes.Length;
			for (int i = 0; i < length; ++i)
				this.AddShape(shapes[i]);
		}


		/// <summary>
		/// Add Shapes to this Shape
		/// </summary>
		/// <param name="shapes">The Shapes</param>
		public void AddShapes(List<Shape> shapes)
		{
			int length = shapes.Count;
			for (int i = 0; i < length; ++i)
				this.AddShape(shapes[i]);
		}


		#endregion

        #region ADD CONNECTED SHAPE

        /// <summary>
        /// Adds the following shape to the list of connected shapes attached to the correct side with respect to orientation
        /// </summary>
        /// <param name="s"></param>
        public void addShape(Shape s, Orientation o)
        {
            XmlAttrs.Connections.Add(s, "connected");
            connectedShapes.Add(s, "connected");
            if (o == Orientation.Down)
                directionallyConnectedShapes[DOWN].Add(s);
            if (o == Orientation.Left)
                directionallyConnectedShapes[LEFT].Add(s);
            if (o == Orientation.Right)
                directionallyConnectedShapes[RIGHT].Add(s);
            if (o == Orientation.Up)
                directionallyConnectedShapes[UP].Add(s);


        }



        /// <summary>
        /// This is a temporary function that is only used while we do not 
        /// distinguish between input and output connections
        /// </summary>
        /// <param name="s"></param>
        public void addConnectedShape(Shape s, string connectionType)
        {
            if(!XmlAttrs.Connections.ContainsKey(s))
                XmlAttrs.Connections.Add(s, connectionType);
            if(!connectedShapes.ContainsKey(s))
                connectedShapes.Add(s, connectionType);
        }

        public void removeConnectedShape(Shape s)
        {
            if (XmlAttrs.Connections.ContainsKey(s))
                XmlAttrs.Connections.Remove(s);
            if(connectedShapes.ContainsKey(s))
                connectedShapes.Remove(s);
        }

        #endregion

        #endregion

        #region REMOVE FROM SHAPE

        #region REMOVE SUBSTROKE(S)

        /// <summary>
		/// Removes a Substroke from the Shape.
		/// </summary>
		/// <param name="substroke">Substroke to remove</param>
		/// <returns>True iff the Substroke is removed</returns>
		public bool RemoveSubstroke(Substroke substroke)
		{
			//Console.WriteLine(this.XmlAttrs.Id + ": DEL :" + substroke.XmlAttrs.Id);
			if (this.substrokes.Contains(substroke))
			{
				substroke.ParentShapes.Remove(this);
				this.substrokes.Remove(substroke);

				ReapIfEmpty();

				UpdateAttributes();
               
				
				return true;
			}
			else
			{
				return false;
			}
		}

        /// <summary>
        /// Removes a Substroke from the Shape by its Id.
        /// </summary>
        /// <param name="substroke">Substroke to remove</param>
        /// <returns>True iff the Substroke is removed</returns>
        public bool RemoveSubstrokeByID(Substroke substroke)
        {
            //Console.WriteLine(this.XmlAttrs.Id + ": DEL :" + substroke.XmlAttrs.Id);
            foreach (Substroke substroke1 in substrokes)
            {
                if (substroke1.Id == substroke.Id)
                {
                    substroke1.ParentShapes.Remove(this);
                    substrokes.Remove(substroke1);
                    ReapIfEmpty();

                    UpdateAttributes();

                    return true;
                }
            }
            return false;
        }

        /// <summary>
        /// Removes an ArrayList of Substrokes from the Shape by ID.
        /// </summary>
        /// <param name="substrokes">Substrokes to remove</param>
        /// <returns>True iff all Substrokes are removed</returns>
        public bool RemoveSubstrokesByID(List<Substroke> substrokes)
        {
            bool completelyRemoved = true;

            Substroke currSubstroke;
            int len = substrokes.Count;
            for (int i = 0; i < len; ++i)
            {
                currSubstroke = substrokes[i];

                if (!RemoveSubstrokeByID(currSubstroke))
                {
                    completelyRemoved = false;
                    Console.WriteLine("Substroke " + currSubstroke.XmlAttrs.Id.Value + " not removed!");
                }
            }

            return completelyRemoved;
        }

		
		/// <summary>
		/// Removes an ArrayList of Substrokes from the Shape.
		/// </summary>
		/// <param name="substrokes">Substrokes to remove</param>
		/// <returns>True iff all Substrokes are removed</returns>
		public bool RemoveSubstrokes(List<Substroke> substrokes)
		{
			bool completelyRemoved = true;

            Substroke currSubstroke;
            int len = substrokes.Count;
			for (int i = 0; i < len; ++i)
			{
				currSubstroke = substrokes[i];

				if (!RemoveSubstroke(currSubstroke))
				{
					completelyRemoved = false;
					Console.WriteLine("Substroke " + currSubstroke.XmlAttrs.Id.Value + " not removed!");
				}
			}

			return completelyRemoved;
		}

		#endregion

		#region REMOVE SHAPE(S)

		/// <summary>
		/// Removes a Shape from the Shape.
		/// </summary>
		/// <param name="shape">Shape to remove</param>
		/// <returns>True iff the Shape is removed</returns>
		public bool RemoveShape(Shape shape)
		{
			if (this.shapes.Contains(shape))
			{
				shape.ParentShape = null;
				this.shapes.Remove(shape);
				UpdateAttributes();
				
				return true;
			}
			else
			{
				return false;
			}
		}

		
		/// <summary>
		/// Removes an ArrayList of Shapes from the Shape.
		/// </summary>
		/// <param name="shapes">Shapes to remove</param>
		/// <returns>True iff all Shapes are removed</returns>
		public bool RemoveShapes(List<Shape> shapes)
		{
			bool completelyRemoved = true;

            Shape currShape;
            int len = shapes.Count;
			for (int i = 0; i < len; ++i)
			{
				currShape = shapes[i];

				if (!RemoveShape(currShape))
				{
					completelyRemoved = false;
					Console.WriteLine("Shape " + currShape.XmlAttrs.Id.Value + " not removed!");
				}
			}

			return completelyRemoved;
		}

		#endregion

		#endregion

		#region UPDATE ATTRIBUTES

		/// <summary>
		/// Updates the spatial attributes of the Shape, such as the origin
		/// and width of the shape
		/// </summary>
		public void UpdateAttributes()
		{
            float minX = Single.PositiveInfinity;
            float maxX = Single.NegativeInfinity;

			float minY = Single.PositiveInfinity;
			float maxY = Single.NegativeInfinity;

			// Cycle through the Substrokes within the Shape
			foreach (Substroke s in substrokes)
			{
                s.UpdateAttributes();
				minX = Math.Min(minX, s.XmlAttrs.X.Value);
				minY = Math.Min(minY, s.XmlAttrs.Y.Value);

				maxX = Math.Max(maxX, s.XmlAttrs.X.Value + s.XmlAttrs.Width.Value);
				maxY = Math.Max(maxY, s.XmlAttrs.Y.Value + s.XmlAttrs.Height.Value);
			}

			// Cycle through the other Shapes within the Shape
			foreach (Shape sh in shapes)
			{
				if (sh.XmlAttrs.X == null || sh.XmlAttrs.Y == null ||
					sh.XmlAttrs.Width == null || sh.XmlAttrs.Height == null)
				{
					// Since sh is an object, it gets updated when this is called
					sh.UpdateAttributes();
				}

				minX = Math.Min(minX, sh.XmlAttrs.X.Value);
				minY = Math.Min(minY, sh.XmlAttrs.Y.Value);

				maxX = Math.Max(maxX, sh.XmlAttrs.X.Value + sh.XmlAttrs.Width.Value);
				maxY = Math.Max(maxY, sh.XmlAttrs.Y.Value + sh.XmlAttrs.Height.Value);
			}

			// Set the origin at the top-left corner of the shape group
			this.XmlAttrs.X = minX;
			this.XmlAttrs.Y = minY;
            

			// Set the width and height of the shape
			this.XmlAttrs.Width = maxX - minX;
			this.XmlAttrs.Height = maxY - minY;

			// Sort the substrokes to ensure we are still in an ascending time order
			this.substrokes.Sort();

			// Update the Start, End and Time attributes
			if (this.substrokes.Count > 0)
			{
				this.XmlAttrs.Start = this.substrokes[0].XmlAttrs.Id;
				this.XmlAttrs.End = this.substrokes[this.substrokes.Count - 1].XmlAttrs.Id;
				this.XmlAttrs.Time = this.substrokes[this.substrokes.Count - 1].XmlAttrs.Time;
			}
			else
			{
				this.XmlAttrs.Start = null;
				this.XmlAttrs.End = null;
				this.XmlAttrs.Time = null;
			}

			if (this.shapes.Count > 0 && 
				this.shapes[0].XmlAttrs.Time.Value < this.XmlAttrs.Time.Value)
			{
				this.XmlAttrs.Time = this.shapes[0].XmlAttrs.Time;
			}
		}

		/// <summary>
		/// Update the attributes of the given shape
		/// </summary>
		/// <param name="s">The shape to update the attributes of</param>
		public static void UpdateAttributes(Shape s)
		{
			s.UpdateAttributes();
		}

		private void ReapIfEmpty()
		{
			if (this.substrokes.Count == 0 && this.shapes.Count == 0)
			{
				// Do some reaping here...
			}
		}

		#endregion

		#region GETTERS & SETTERS

        public Orientation Orient
        {
            get { return this.orient; }
            set { this.orient = value; }
        }

        public List<Shape> DownShapes
        {
            get { return directionallyConnectedShapes[DOWN]; }
        }
        public List<Shape> LeftShapes
        {
            get { return directionallyConnectedShapes[LEFT]; }
        }
        public List<Shape> RightShapes
        {
            get { return directionallyConnectedShapes[RIGHT]; }
        }
        public List<Shape> UpShapes
        {
            get { return directionallyConnectedShapes[UP]; }
        }

        public Dictionary<Shape, string>ConnectedShapes
        {
            get { return connectedShapes; }
        }


		/// <summary>
		/// Mysterious index of this shape
		/// </summary>
		public int Index
		{
			get
			{
				return index;
			}
			set
			{
				index = value;
			}
		}

		/// <summary>
		/// Get/Set the Label on a given Shape
		/// </summary>
		public string Label
		{
			get
			{
				return XmlAttrs.Type;
			}
			set
			{
				XmlAttrs.Type = value;
			}
		}

		/// <summary>
		/// Get/Set the lower-case Label on a given Shape
		/// </summary>
		public string LabelL
		{
			get
			{
				return XmlAttrs.Type.ToLower();
			}
			set
			{
				XmlAttrs.Type = value.ToLower();
			}
		}

		/// <summary>
		/// Get the Probability of the Label on a given Shape
		/// </summary>
		public float Probability
		{
			get
			{
				if (XmlAttrs.Probability == null)
				{
					return 0.0F;
				}
				else
				{
					return (float)XmlAttrs.Probability;
				}
			}
            set
            {
                XmlAttrs.Probability = value;
            }
		}

		/// <summary>
		/// Get Substrokes
		/// </summary>
		public Substroke[] Substrokes
		{
			get
			{
				return this.substrokes.ToArray();
			}
		}

		/// <summary>
		/// Get a List of Substrokes
		/// </summary>
        public List<Substroke> SubstrokesL
        {
            get
            {
                return this.substrokes;
            }
        }

		/// <summary>
		/// Get Shapes
		/// </summary>
		public Shape[] Shapes
		{
			get
			{
				return this.shapes.ToArray();
			}
		}

		/// <summary>
		/// Get a List of Shapes
		/// </summary>
        public List<Shape> ShapesL
        {
            get
            {
                return this.shapes;
            }
        }


		/// <summary>
		/// Get or set ParentShape
		/// </summary>
		public Shape ParentShape
		{
			get
			{
				if (this.parentShape == null)
					return new Shape();
				else
                    return this.parentShape;
			}

			set
			{
				this.parentShape = value;
			}
		}

		/// <summary>
		/// The weighted centroid of this shape. Returns coordinates of the form [x, y].
		/// </summary>
		public double[] Centroid
		{
			get
			{
				double[] centroid = new double[2];
				int count = 0;
				foreach (Substroke ss in substrokes)
				{
					foreach (Point p in ss.PointsL)
					{
						centroid[0] += p.X;
						centroid[1] += p.Y;
						++count;
					}
				}
				centroid[0] /= count;
				centroid[1] /= count;
				return centroid;
			}
		}

		/// <summary>
		/// The GUID of this shape
		/// </summary>
		public Guid Id
		{
			get
			{
				if (XmlAttrs.Id == null)
					XmlAttrs.Id = new Guid();
				return (Guid)XmlAttrs.Id;
			}
			set
			{
				XmlAttrs.Id = value;
			}
		}

		/// <summary>
		/// Is this shape a digital logic gate or not?
		/// </summary>
		public bool IsGate
		{
			get
			{
				if (LabelL == "and" ||
					LabelL == "not" ||
					LabelL == "or" ||
					LabelL == "nor" ||
					LabelL == "nand" ||
					LabelL == "bubble" ||
					LabelL == "notbubble" ||
					LabelL == "xor" ||
					LabelL == "xnor" ||
                    LabelL == "male" ||
                    LabelL == "female" ||
                    LabelL == "labelbox")
					return true;
				return false;
			}
		}

        public List<Point> Points
        {
            get
            {
                List<Point> points = new List<Point>();
                foreach (Substroke sub in this.substrokes)
                {
                    points.AddRange(sub.PointsL);
                }
                return points;
            }
        }

        public String Name
        {
            get { return this.name; }
            set { this.name = value; }
        }

        public String Type
        {
            get { return this.XmlAttrs.Type; }
            set { this.XmlAttrs.Type = value; }
        }


		#endregion

		#region OTHER

		/// <summary>
		/// Change this parent shape
		/// </summary>
		/// <param name="shape"></param>
		public void ChangeParentShape(Shape shape)
		{
			this.parentShape = shape;
		}


		/// <summary>
		/// Compute the Clone of this shape.
		/// </summary>
		/// <returns>The Clone of this shape.</returns>
		public Shape Clone()
		{
			return new Shape(this);
		}


		/// <summary>
		/// Compare this Shape to another based on time.
		/// Returns less than 0 if this time is less than the other's.
		/// Returns 0 if this time is equal to the other's.
		/// Returns greater than if this time is greater than the other's.
		/// </summary>
		/// <param name="obj">The other Shape to compare this one to</param>
		/// <returns>An integer indicating how the Shape times compare</returns>
		int System.IComparable.CompareTo(Object obj)
		{
			return (int)(this.XmlAttrs.Time.Value - ((Stroke)obj).XmlAttrs.Time.Value);
		}

		/// <summary>
		/// Overrides the Equality of Shapes to depend on whether two shapes' substrokes are equal
		/// </summary>
		/// <param name="other">The other shape the compare to</param>
		/// <returns>Whether or not the Shapes are equal</returns>
		public bool Equals(Shape other)
		{
			if (substrokes.Count == other.substrokes.Count)
			{
				substrokes.Sort(new SubstrokeGuidComparer());
				other.substrokes.Sort(new SubstrokeGuidComparer());
				for (int i = 0; i < substrokes.Count; ++i)
				{
					if (!substrokes[i].Equals(other.substrokes[i]))
						return false;
				}
				substrokes.Sort();
				other.substrokes.Sort();
				return true;
			}
			return false;
		}

		/// <summary>
		/// Scale the shape along the x-axis
		/// </summary>
		/// <param name="scaleFactor">Scale factor</param>
		/// <param name="xStart">X point</param>
		/// <param name="yStart">Y point</param>
		public void scaleX(double scaleFactor, double xStart, double yStart)
		{
			for (int i = 0; i < substrokes.Count; ++i)
			{
				substrokes[i].stretchX(scaleFactor, xStart, yStart);
			}
		}

        #endregion
    }
}
