/*
 * File: Substroke.cs
 *
 * Authors: Aaron Wolin, Devin Smith, Jason Fennell, and Max Pflueger.
 * Harvey Mudd College, Claremont, CA 91711.
 * Sketchers 2006.
 * 
 * Use at your own risk.  This code is not maintained and not guaranteed to work.
 * We take no responsibility for any harm this code may cause.
 */

using System;
using System.Collections.Generic;

namespace Sketch
{
	/// <summary>
	/// Substroke class.
	/// </summary>
	[Serializable]
	public class Substroke : IComparable<Substroke>
	{
		#region INTERNALS

		/// <summary>
		/// The points of the Substroke
		/// </summary>
        private List<Point> points;
		
		/// <summary>
		/// This is the parent stroke
		/// </summary>
		private Stroke parentStroke;
        
		/// <summary>
		/// This is the parent shape
		/// </summary>
        private List<Shape> parentShapes;

		/// <summary>
		/// The XML attributes of the Substroke
		/// </summary>
		public XmlStructs.XmlShapeAttrs XmlAttrs;

		/// <summary>
		/// Spatial Length of the substroke
		/// </summary>
		private double? _spatial_length = null;

        /// <summary>
        /// The result of the Rubine Recognizer
        /// </summary>
        private string strokeType;

		#endregion

		#region CONSTRUCTORS

		/// <summary>
		/// Constructor
		/// </summary>
		public Substroke() 
			: this(new List<Point>(), new XmlStructs.XmlShapeAttrs(true))
		{
			// Calls the main constructor
		}
		
		/// <summary>
		/// Creates a Substroke from an array of Points and the XML attributes
		/// </summary>
		/// <param name="points">Array of Points</param>
		/// <param name="XmlAttrs">The XML attributes of the stroke</param>
		public Substroke(Point[] points, XmlStructs.XmlShapeAttrs XmlAttrs)
			: this(new List<Point>(points), XmlAttrs)
		{
            
		}
		
		/// <summary>
		/// Creates a Substroke from a List of Points and the XML attributes
		/// </summary>
		/// <param name="points">List of Points</param>
		/// <param name="XmlAttrs">The XML attributes of the stroke</param>
		public Substroke(List<Point> points, XmlStructs.XmlShapeAttrs XmlAttrs)
		{           
            this.points = new List<Point>(points.Count);
            this.parentShapes = new List<Shape>();
			this.parentStroke = null;
			this.XmlAttrs = XmlAttrs;

			this.AddPoints(points);
            this.UpdateAttributes();
		}
		
		
		#endregion
		
		#region ADD TO SUBSTROKE
		
		/// <summary>
		/// Adds the Point to the Substroke's Point List. Uses binary search to find insertion spot
		/// </summary>
		/// <param name="point">Point to add</param>
		public void AddPoint(Point point)
		{
			//if (this.points.Contains(point))
			//	this.points.Remove(point);

			if (!this.points.Contains(point))
			{
				int low = 0;
				int high = this.points.Count - 1;
				int mid;
				while (low <= high)
				{
					mid = (high - low) / 2 + low;
					if (point.Time < points[mid].Time)
						high = mid - 1;
					else
						low = mid + 1;
				}
		
				this.points.Insert(low, point);

				// Reset cached length
				_spatial_length = null;
			}
		}

        /// <summary>
        /// For some reason, we know this point goes on the end of the list
        /// and we trust that it's not already in the substroke.
        /// </summary>
        /// <param name="point"></param>
        public void AppendPoint(Point point)
        {
            this.points.Add(point);
            _spatial_length = null;
        }


		/// <summary>
		/// Adds the Points to the Substroke's Point ArrayList.
		/// </summary>
		/// <param name="points">Points to add</param>
		public void AddPoints(Point[] points)
		{
			int length = points.Length;
			
			for (int i = 0; i < length; i++)
				this.AddPoint(points[i]);
		}

		
		/// <summary>
		/// Adds the Points to the Substroke's Point ArrayList.
		/// </summary>
		/// <param name="points">Points to add</param>
		public void AddPoints(List<Point> points)
		{
			int length = points.Count;
			
			for (int i = 0; i < length; i++)
				this.AddPoint(points[i]);
            //TODO! Fix this line, for some reason it gives an index out of range exception
			//this.XmlAttrs.Time = points[points.Count - 1].XmlAttrs.Time;
		}

		/// <summary>
		/// Add a group of substrokes to this substroke. They should be such that it makes sense to add them (i.e., adjacent)
		/// </summary>
		/// <param name="sss">The substrokes to add</param>
		public void AddSubstrokes(List<Substroke> sss)
		{
            points.Sort();
			for(int i = 0; i < sss.Count; ++i)
			{
				AddSubstroke(sss[i], false);
			}
            UpdateAttributes();
		}

		/// <summary>
		/// Combine this substroke with another substroke
		/// </summary>
		/// <param name="ss">The substroke to combine with</param>
        public void AddSubstroke(Substroke ss)
        {
            AddSubstroke(ss, true);
        }

		/// <summary>
		/// Combines this with another substroke
		/// </summary>
		/// <param name="ss">The substroke to combine with</param>
		/// <param name="update">Update the properties of the substroke?</param>
		private void AddSubstroke(Substroke ss, bool update)
		{
            if (update)
            {
                points.Sort();
                ss.points.Sort();
            }

			//Invalidate properties
			XmlAttrs.Height = null;
			XmlAttrs.Width = null;
			XmlAttrs.Area = null;
			XmlAttrs.LeftX = null;
			XmlAttrs.TopY = null;
			XmlAttrs.X = null;
			XmlAttrs.Y = null;

			//Add the points
			AddPoints(ss.points);
			ss.points.RemoveRange(0, ss.points.Count);
			XmlAttrs.End = points[points.Count - 1].XmlAttrs.Id;
			XmlAttrs.Time = points[points.Count - 1].XmlAttrs.Time;
			if (update)
                UpdateAttributes();
            
            if (ss.parentStroke != null)
			    ss.parentStroke.RemoveSubstroke(ss);

			//The substrokes now need to share the same ParentShapes
            if (ss.parentShapes != null)
            {
                for (int i = 0; i < ss.parentShapes.Count; ++i)
                {
                    Shape ps = ss.parentShapes[i];
                    if (parentShapes.Contains(ps))
                        continue;
                    ps.AddSubstroke(this);
                }
                ss.removeFromParentShapes();
            }
		}

		#endregion

		#region REMOVE

		/// <summary>
		/// Deletes a range of Points from the Substroke.
		/// </summary>
		/// <param name="index">Index of the initial Point to delete</param>
		/// <param name="count">How many to remove</param>
		public void DeleteRange(int index, int count)
		{
			this.points.RemoveRange(index, count);

			// Reset cached length
			_spatial_length = null;
		}


		/// <summary>
		/// Removes a label from the Substroke
		/// </summary>
		/// <param name="label">Label to remove</param>
		/// <returns>The shape that was removed</returns>
		public Shape RemoveLabel(String label)
		{
			foreach (Shape shape in this.parentShapes)
			{
				if (shape.XmlAttrs.Type.Equals(label))
				{
					shape.RemoveSubstroke(this);
					return shape;
				}
			}

			return null;
		}

		#endregion

        /// <summary>
        /// Updates X, Y, Width, Height, Time
        /// </summary>
        public void UpdateAttributes()
        {
            float minX = Single.PositiveInfinity;
            float maxX = Single.NegativeInfinity;

            float minY = Single.PositiveInfinity;
            float maxY = Single.NegativeInfinity;

            Point point;
            int len = points.Count;
            int i;
            for (i = 0; i < len; ++i)
            {
                point = points[i];

                minX = Math.Min(minX, point.X);
                minY = Math.Min(minY, point.Y);

                maxX = Math.Max(maxX, point.X);
                maxY = Math.Max(maxY, point.Y);
            }

            this.XmlAttrs.X = minX;
            this.XmlAttrs.Y = minY;
            this.XmlAttrs.Width = maxX - minX;
            this.XmlAttrs.Height = maxY - minY;

            //TODO!  Fix this line, for some reason it gives an index out of range exception
			//this.XmlAttrs.Time = points[points.Count - 1].XmlAttrs.Time;

            //Other attributes are updated in split
        }

        #region SPLITTERS

        /// <summary>
        /// Splits a Substroke at the given index.
        /// </summary>
        /// <param name="index">Index where the Substroke should be split</param>
        public void SplitAt(int index)
		{
			// If there is nothing to split...
			if (index < 1 || index > this.points.Count - 2)
				return;
			
			Substroke lastHalf = new Substroke();

			// We invalidate these things
			this.XmlAttrs.Height = null;
			this.XmlAttrs.Width  = null;
			this.XmlAttrs.Area   = null;
			this.XmlAttrs.LeftX  = null;
			this.XmlAttrs.TopY   = null;
			this.XmlAttrs.X      = null;
			this.XmlAttrs.Y      = null;

			// Copy the rest of the attributes into the second half
			lastHalf.XmlAttrs = this.XmlAttrs.Clone();
			lastHalf.XmlAttrs.Id = System.Guid.NewGuid();

			int length = points.Count;

			// Add points to the last Half
			lastHalf.AddPoints(points.GetRange(index, length - index)); 

			// Change the start of the last half
			lastHalf.XmlAttrs.Start = lastHalf.points[0].XmlAttrs.Id;

			// Remove all the points in the second half
			this.points.RemoveRange(index, length - index);
			//this.points.RemoveRange(index + 1, length - index - 1);	// Used if we want to include the middle, split point in both Substrokes
			
			// Add the end to the first half
			this.XmlAttrs.End = this.points[index - 1].XmlAttrs.Id;

			// The new first half time is equal to the last point's time in it
			this.XmlAttrs.Time = this.points[index - 1].XmlAttrs.Time;

            //Update X, Y...
            this.UpdateAttributes();
            lastHalf.UpdateAttributes();

			// Update the parent info
			lastHalf.ParentStroke = this.parentStroke;
			lastHalf.ParentStroke.AddSubstroke(lastHalf);

            //lastHalf.ParentShapes = (ArrayList)this.parentShapes;
            lastHalf.ParentShapes = this.parentShapes.GetRange(0, this.parentShapes.Count); //Like clone
			
			length = lastHalf.parentShapes.Count;
			for (int i = 0; i < length; ++i)
				if (lastHalf.parentShapes[i] != null)
					lastHalf.parentShapes[i].AddSubstroke(lastHalf);

		}


		/// <summary>
		/// Splits a Substroke at the given indices.
		/// </summary>
		/// <param name="indices">Indices where the Substroke should be split</param>
		public void SplitAt(int[] indices)
		{
			// Due to the way that we insert substrokes into the array, we must step through backwards...
			Array.Sort(indices);
			Array.Reverse(indices);

			int length = indices.Length;
			for (int i = 0; i < length; i++)
				this.SplitAt(indices[i]);
		}

		#endregion

		#region GETTERS & SETTERS
		
		/// <summary>
		/// Get the number of Points in this substroke
		/// </summary>
		public int Length
		{
			get
			{
				return this.points.Count;
			}
		}

		/// <summary>
		/// Returns the spatial length of this substroke
		/// </summary>
		public double SpatialLength
		{
			get
			{
				if (_spatial_length == null)
					calculateSpatialLength();
				return (double)_spatial_length;
			}
		}

		/// <summary>
		/// Gets a Point[] of the points contained in the Substroke.
		/// </summary>
		public Point[] Points
		{
			get
			{
				return points.ToArray();
			}
		}

        /// <summary>
        /// Gets a Point List of the points contained in the Substroke
        /// </summary>
        public List<Point> PointsL
        {
            get
            {
                return points;
            }
        }

		/// <summary>
		/// Get the Points in this substroke as an array of System.Drawing.PointFs.
		/// </summary>
		public System.Drawing.PointF[] PointsAsPointFs
		{
			get
			{
				System.Drawing.PointF[] pts = new System.Drawing.PointF[points.Count];
				for (int i = 0; i < points.Count; ++i)
					pts[i] = points[i].SysDrawPointF;
				return pts;
			}
		}

        /// <summary>
        /// Get the Points in this substroke as an array of System.Drawing.Points.
        /// </summary>
        public System.Drawing.Point[] PointsAsSysPoints
        {
            get
            {
                System.Drawing.Point[] pts = new System.Drawing.Point[points.Count];
                for (int i = 0; i < points.Count; ++i)
                    pts[i] = points[i].SysDrawPoint;
                return pts;
            }
        }

			/// <summary>
			/// Get or set ParentStroke
			/// </summary>
			public Stroke ParentStroke
		{
			get
			{
				return this.parentStroke;
			}

			set
			{
				this.parentStroke = value;
			}
		}


		/// <summary>
		/// Get or set ParentShapes
		/// </summary>
		public List<Shape> ParentShapes
		{
			get
			{
                return this.parentShapes;
			}

			set
			{
                this.parentShapes = value;
			}
		}


		/// <summary>
		/// Get the labels associated with a Substroke.
		/// </summary>
		/// <returns></returns>
		public string[] Labels
		{
			get
			{
				List<string> labels = new List<string>();

				Shape shape;
				int length = this.ParentShapes.Count;
				for (int i = 0; i < length; ++i)
				{
					shape = this.ParentShapes[i];

					if (!labels.Contains(shape.XmlAttrs.Type))
						labels.Add(shape.XmlAttrs.Type);
				}

				return labels.ToArray();
			}
		}


		/// <summary>
		/// Get the beliefs associated with the labels 
		/// </summary>
		/// <returns></returns>
		public double[] GetBeliefs()
		{
			List<string> labels = new List<string>();
			List<double> beliefs = new List<double>();

			Shape shape;
			int length = this.ParentShapes.Count;
			for (int i = 0; i < length; ++i)
			{
				shape = this.ParentShapes[i];
				
				if (!labels.Contains(shape.XmlAttrs.Type))
				{
					labels.Add(shape.XmlAttrs.Type);
					beliefs.Add(shape.XmlAttrs.Probability.Value);
				}
			}
			
			return beliefs.ToArray();
		}


		/// <summary>
		/// Get the first label
		/// </summary>
		/// <returns></returns>
		public string FirstLabel
		{
			get
			{
				if (this.parentShapes.Count == 0)
					return "unlabeled";
				else
					return this.parentShapes[0].Label;
			}
		}

        /// <summary>
        /// Get the first label in lower case
        /// </summary>
        public string FirstLabelL
        {
            get
            {
                if (this.parentShapes.Count == 0)
                    return "unlabeled";
                else return parentShapes[0].LabelL;
            }
        }

		/// <summary>
		/// Get the first probability
		/// </summary>
		/// <returns></returns>
		public double GetFirstBelief()
		{
			if(this.parentShapes == null || this.parentShapes.Count == 0)
				return -1.0;
			else
			{
                if (!this.parentShapes[0].XmlAttrs.Probability.HasValue)
                    return -1.0;
                else
                    return this.ParentShapes[0].XmlAttrs.Probability.Value;
			}
		}

		/// <summary>
		/// The GUID associated with this substroke
		/// </summary>
		public Guid Id
		{
			get
			{
				if (XmlAttrs.Id == null)
					XmlAttrs.Id = new Guid();
				return (Guid)XmlAttrs.Id;
			}
			set
			{
				XmlAttrs.Id = value;
			}
		}

        /// <summary>
        /// Find the centroid for the substroke
        /// returns a double array {X,Y}
        /// </summary>
        public double[] Centroid
        {
            get
            {
                double[] res = new double[] { 0d, 0d };

                foreach (Point p in PointsL)
                {
                    res[0] += p.X;
                    res[1] += p.Y;
                }

                res[0] /= PointsL.Count;
                res[1] /= PointsL.Count;

                return res;
            }
        }
		/// <summary>
		/// Find the two endpoints of this sketch. Tolerant of overstroking
		/// </summary>
		public Point[] Endpoints
		{
			get
			{
				// TODO: Find a better way to calculate endpoints
				return new Point[2] { points[0], points[points.Count - 1] };
			}
		}

        /// <summary>
        /// This is the classification of the substroke as gate, label, or wire
        /// </summary>
        public String Classification
        {
            get 
            { 
                if (this.XmlAttrs.Classification == null)
                    return "Unknown";
                return this.XmlAttrs.Classification; }
            set 
            { 
                this.XmlAttrs.Classification = value;
                if (this.strokeType == null)
                    this.strokeType = this.XmlAttrs.Classification;
            }
        }

        /// <summary>
        /// This is the probability associated with the classification
        /// </summary>
        public float ClassificationBelief
        {
            get 
            {
                if (this.XmlAttrs.ClassificationBelief == null)
                    return 0f;
                else
                    return (float)this.XmlAttrs.ClassificationBelief;
            }
            set { this.XmlAttrs.ClassificationBelief = value; }
        }
        /// <summary>
        /// The type of the stroke as classified by the Rubine Recognizer
        /// </summary>
        public string StrokeType
        {
            get { return this.strokeType; }
            set { this.strokeType = value; }
        }

		#endregion

		#region OTHER

		/// <summary>
		/// Calculate the cached spatial length
		/// </summary>
		private void calculateSpatialLength()
		{
			double dist = 0.0;
			List<Point> ps = PointsL;
			for (int i = 0; i < ps.Count - 1; ++i)
			{
				dist += ps[i].distance(ps[i + 1]);
			}
			_spatial_length = dist;
		}

		/// <summary>
		/// Removes a Substroke from all of its parent shapes
		/// </summary>
        public void removeFromParentShapes()
        {
			List<Shape> ps = new List<Shape>(parentShapes);
            for (int i = 0; i < ps.Count; ++i)
            {
                ps[i].RemoveSubstroke(this);
            }
        }

        /// <summary>
        /// Compares Substrokes for Equality on the basis of their GUIDs.
        /// </summary>
        /// <param name="substroke"></param>
        /// <returns></returns>
        public bool Equals(Substroke substroke)
        {
            return this.XmlAttrs.Id.Equals(substroke.XmlAttrs.Id);
        }

        /// <summary>
        /// Returns the Hash Code of this Substroke
        /// </summary>
        /// <returns></returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

		/// <summary>
		/// Compare this Substroke to another based on time.
		/// Returns less than 0 if this time is less than the other's.
		/// Returns 0 if this time is equal to the other's.
		/// Returns greater than 0 if this time is greater than the other's.
		/// </summary>
		/// <param name="substroke">The other Substroke to compare this one to</param>
		/// <returns>An integer indicating how the Substroke times compare</returns>
        int System.IComparable<Substroke>.CompareTo(Substroke substroke)
        {
            return (int)(this.XmlAttrs.Time.Value - substroke.XmlAttrs.Time.Value);
        }

        /// <summary>
        /// Clones a Substroke (partial-deep copy. Sort of silly as it stands now).
        /// </summary>
        /// <returns>A clone of the current substroke</returns>
        public Substroke Clone()
        {
            Substroke clone = new Substroke(this.ClonePoints(), this.XmlAttrs.Clone());
            
            //Do we want to deep clone parent shapes or should they be left as references?
            if (this.parentShapes != null)
                clone.ParentShapes = this.CloneParentShapes();

            //Do we want to deep clone parent stroke or should they be left as references?
            if (this.parentStroke != null)
                clone.ParentStroke = this.parentStroke.Clone();
            return clone;
        }

        /// <summary>
        /// Returns a deep copy of the points
        /// </summary>
        /// <returns></returns>
        internal List<Point> ClonePoints()
        {
            int i, len = this.points.Count;
            List<Point> copyPoints = new List<Point>(len);
            for (i = 0; i < len; ++i)
                copyPoints.Add(this.points[i].Clone());
            return copyPoints;
        }

        /// <summary>
        /// Returns a shallow copy of the points
        /// </summary>
        /// <returns></returns>
        internal List<Shape> CloneParentShapes()
        {
            return new List<Shape>(this.parentShapes);
            /*
            int i, len = this.parentShapes.Count;
            List<Shape> copyShapes = new List<Shape>(len);
            for (i = 0; i < len; ++i)
                copyShapes.Add(this.parentShapes[i].Clone());
            return copyShapes;
            */
        }


		/// <summary>
		/// Clone the substroke, but empty its parentShapes and parentStrokes
		/// </summary>
		/// <returns>The clone of the Substroke</returns>
		internal Substroke CloneConstruct()
		{
			return new Substroke(ClonePoints(), this.XmlAttrs.Clone());
		}

		/// <summary>
		/// Return a string representation of this substroke
		/// </summary>
		/// <returns></returns>
		public override string ToString()
		{
			return String.Format("Substroke {0}", Id);
		}


		/// <summary>
		/// Resample this substroke to contain the given number of points and return the
		/// resampled substroke (note: does not modify the current object)
		/// </summary>
		/// <param name="PPS">The number of points to resize to</param>
		/// <returns>The new substroke</returns>
		public Substroke Resample(int PPS)
		{
			List<Point> spc = new List<Point>();

			double I = SpatialLength / (PPS - 1);
			double D = 0;

			spc.Add(points[0]);

			for (int i = 1; i < points.Count; i++)
			{
				double d = points[i - 1].distance(points[i]);
				if (D + d >= I)
				{
					Point q = new Point(points[i - 1]);
                    q.XmlAttrs.Id = Guid.NewGuid();

					q.X += (float)(((I - D) / d) * (points[i].X - points[i - 1].X));
					q.Y += (float)(((I - D) / d) * (points[i].Y - points[i - 1].Y));
					spc.Add(q);

					points.Insert(i, q);
					D = 0.0;
				}
				else
				{
					D += d;
				}
			}

			if (spc.Count < PPS) spc.Add(points[points.Count - 1]);
            if (spc.Count != PPS) throw new Exception("Stroke resampling failed");
			Substroke res = new Substroke(spc, new XmlStructs.XmlShapeAttrs(true));
			res.XmlAttrs.Name = "substroke";
			res.XmlAttrs.Type = "substroke";
			res.XmlAttrs.Time = XmlAttrs.Time;
			res.UpdateAttributes();
            //if (res.PointsL.Count != PPS) throw new Exception("we lost a point somewhere!");
			return res;
		}

		/// <summary>
		/// Resample this substroke to contain the given number of points
		/// </summary>
		/// <param name="PPS">The new number of points to contain</param>
		public void ResampleInPlace(int PPS)
		{
			List<Point> spc = new List<Point>();

			double I = SpatialLength / (PPS - 1);
			double D = 0;

			spc.Add(points[0]);

			for (int i = 1; i < points.Count; i++)
			{
				double d = points[i - 1].distance(points[i]);
				if (D + d >= I)
				{
					Point q = new Point(points[i - 1]);

					q.X += (float)(((I - D) / d) * (points[i].X - points[i - 1].X));
					q.Y += (float)(((I - D) / d) * (points[i].Y - points[i - 1].Y));
					spc.Add(q);

					points.Insert(i, q);
					D = 0.0;
				}
				else
				{
					D += d;
				}
			}

			if (spc.Count < PPS) spc.Add(points[points.Count - 1]);

			points = spc;
			UpdateAttributes();
			_spatial_length = null;
		}

        /// <summary>
        /// Returns a reversed copy of this substroke.
        /// </summary>
        /// <returns></returns>
        public Substroke GetReversed()
        {
            Substroke res = new Substroke();
            res.XmlAttrs.Id = new Guid("0000"+this.XmlAttrs.Id.ToString().Substring(4));

            for (int i = 1; i <= points.Count; ++i)
            {
                Point toadd = new Point(points[points.Count - i].X, points[points.Count-i].Y);
                toadd.Time = (ulong)i;
                res.AppendPoint(toadd);
            }

            res.UpdateAttributes();

            return res;
        }

        #endregion

        #region MODIFIERS

        /// <summary>
        /// Rotates a substroke
        /// </summary>
        /// <param name="theta">The angle by which to rotate (in radians)</param>
        /// <param name="xCenter">The center position (x-coord)</param>
        /// <param name="yCenter">The center position (y-coord)</param>
        public void rotate(double theta, double xCenter, double yCenter)
        {
            for (int i = 0; i < points.Count; ++i)
            {
                points[i].rotate(theta, xCenter, yCenter);
            }
        }

		/// <summary>
		/// Create a copy of this object and rotate it around (xCenter, yCenter)
		/// </summary>
		/// <param name="theta">The angle to rotate by, in radians</param>
		/// <param name="xCenter">The x-coordinate of the center position</param>
		/// <param name="yCenter">The y-coordinate of the center position</param>
		/// <returns>The rotated substroke</returns>
        public Substroke cloneRotate(double theta, float xCenter, float yCenter)
        {
            Substroke clone = this.Clone();
            clone.rotate(theta, xCenter, yCenter);
            return clone;
        }

		/// <summary>
		/// Scales a substroke
		/// </summary>
		/// <param name="scaleFactor">The multiplicative amount by which to scale</param>
		/// <param name="xCenter">The point around which we are scaling (x-coord)</param>
		/// <param name="yCenter">The point around which we are scaling (y-coord)</param>
		public void scale(double scaleFactor, double xCenter, double yCenter)
		{
			for (int i = 0; i < points.Count; ++i)
			{
				points[i].scale(scaleFactor, xCenter, yCenter);
			}
			// Reset cached spatial length
			_spatial_length = null;
		}

		/// <summary>
		/// Stretches the substroke along its x-axis
		/// </summary>
		/// <param name="scaleFactor">The amount to scale by</param>
		/// <param name="xStart">X point</param>
		/// <param name="yStart">Y point</param>
		public void stretchX(double scaleFactor, double xStart, double yStart)
		{
			for (int i = 0; i < points.Count; ++i)
			{
				points[i].stretchX(scaleFactor, (float)xStart, (float)yStart);
			}
			_spatial_length = null;
		}

		#endregion

	}
}
