/*
 * File: XmlStructs.cs
 *
 * Authors: Aaron Wolin, Devin Smith, Jason Fennell, and Max Pflueger.
 * Harvey Mudd College, Claremont, CA 91711.
 * Sketchers 2006.
 * 
 * Use at your own risk.  This code is not maintained and not guaranteed to work.
 * We take no responsibility for any harm this code may cause.
 */

using System;
using System.Collections.Generic;

namespace Sketch
{
	/// <summary>
	/// XmlStructs. Contains structs holding variables for relevant XML attributes.
	/// </summary>
	[Serializable]
	public class XmlStructs
	{
		/// <summary>
		/// XML attribute struct for the Sketch data
		/// </summary>
		[Serializable]
		public struct XmlSketchAttrs
		{
			#region INTERNALS

			/// <summary>
			/// The GUID 
			/// </summary>
            public Guid? Id;

			/// <summary>
			/// Units of the Sketch
			/// <example>Units = "Himetric"</example>
			/// </summary>
            public string Units;

            /// <summary>
            /// Name of the sketch
            /// <example>01_EQN1_T</example>
            /// </summary>
            public string Name;

			#endregion

			#region METHODS

			/// <summary>
			/// Get the attribute names.
			/// </summary>
			/// <returns>A string array of attribute names</returns>
			public string[] getAttributeNames()
			{
				return new string[] { "id", "units" };
			}


			/// <summary>
			/// Get the attribute values.
			/// </summary>
			/// <returns>An object array of attribute values</returns>
			public object[] getAttributeValues()
			{
				return new object[] { Id, Units };
			}

			
			/// <summary>
			/// Clone method.
			/// </summary>
			/// <returns>A new XmlSketchAttrs struct copying the old one's values</returns>
			public XmlSketchAttrs Clone()
			{
				XmlSketchAttrs temp = new XmlSketchAttrs();
				temp.Id = this.Id;
				temp.Units = this.Units;
			
				return temp;
			}
			
			#endregion
		}

		
		/// <summary>
		/// XML attribute struct for the Sketcher data
		/// </summary>
		[Serializable]
		public struct XmlSketcherAttrs
		{
			#region INTERNALS

			/// <summary>
			/// GUID of the Sketcher
			/// </summary>
            public Guid? Id;

			/// <summary>
			/// DPI on the X-Axis
			/// </summary>
            public float? XDpi;

			/// <summary>
			/// DPI on the Y-Axis
			/// </summary>
            public float? YDpi;

			/// <summary>
			/// Nickname of the Sketcher
			/// </summary>
            public string Nickname;

			#endregion

			#region METHODS
			
			/// <summary>
			/// Get the attribute names.
			/// </summary>
			/// <returns>A string array of attribute names</returns>
			public string[] getAttributeNames()
			{
				return new string[] { "id", "nickname" };
			}


			/// <summary>
			/// Get the attribute values.
			/// </summary>
			/// <returns>An object array of attribute values</returns>
			public object[] getAttributeValues()
			{
				return new object[] { Id, Nickname };
			}


			/// <summary>
			/// Clone method.
			/// </summary>
			/// <returns>A new XmlSketcherAttrs struct copying the old one's values</returns>
			public XmlSketcherAttrs Clone()
			{
				XmlSketcherAttrs temp = new XmlSketcherAttrs();
				temp.Id = this.Id;
				temp.XDpi = this.XDpi;
				temp.YDpi = this.YDpi;
				temp.Nickname = this.Nickname;

				return temp;
			}
			
			#endregion
		}

		
		/// <summary>
		/// XML attribute struct for the Study data
		/// </summary>
		[Serializable]
		public struct XmlStudyAttrs
		{
			#region INTERNALS

			/// <summary>
			/// Name of the Study
			/// </summary>
			public string Name;

            #endregion

			#region METHODS

			/// <summary>
			/// Clone method.
			/// </summary>
			/// <returns>A new XmlStudyAttrs struct copying the old one's values</returns>
			public XmlStudyAttrs Clone()
			{
				XmlStudyAttrs temp = new XmlStudyAttrs();
				temp.Name = this.Name;

				return temp;
			}
			
			#endregion
		}

		
		/// <summary>
		/// XML attribute struct for the Domain data
		/// </summary>
		[Serializable]
		public struct XmlDomainAttrs
		{
			#region INTERNALS

            /// <summary>
            /// Name of the Domain
            /// </summary>
			public string Name;

			#endregion

			#region METHODS

			/// <summary>
			/// Clone method.
			/// </summary>
			/// <returns>A new XmlDomainAttrs struct copying the old one's values</returns>
			public XmlDomainAttrs Clone()
			{
				XmlDomainAttrs temp = new XmlDomainAttrs();
				temp.Name = this.Name;

				return temp;
			}
			
			#endregion
		}

		
		/// <summary>
		/// XML attribute struct for the Shape data
		/// </summary>
		[Serializable]
		public struct XmlShapeAttrs
		{
			#region DATA

			/// <summary>
			/// GUID of the Shape
			/// </summary>
            public Guid? Id;

			/// <summary>
			/// Name of the Shape
			/// <example>shape</example>
			/// <example>point</example>
			/// </summary>
            public string Name;
			
			/// <summary>
			/// Time of the Shape
			/// </summary>
            public ulong? Time;

			/// <summary>
			/// Type of the Shape
			/// <example>AND</example>
			/// </summary>
            public string Type;

			/// <summary>
			/// Author GUID of the Shape
			/// </summary>
            public Guid? Author;

			/// <summary>
			/// Color of the Shape
			/// </summary>
            public int? Color;

			/// <summary>
			/// Height of the Shape
			/// </summary>
            public float? Height;

			/// <summary>
			/// Width of the Shape
			/// </summary>
            public float? Width;

			/// <summary>
			/// Area (Height x Width) of the Shape
			/// </summary>
            public float? Area;

			/// <summary>
			/// No clue what this does. TODO: Read the MITXML spec
			/// </summary>
            public bool? LaysInk;

			/// <summary>
			/// Orientation of the Shape
			/// </summary>
            public float? Orientation;

			/// <summary>
			/// Pen tip used to draw the Shape
			/// </summary>
            public string PenTip;

			/// <summary>
			/// Are you raster?
			/// </summary>
            public string Raster;

			/// <summary>
			/// What stroke are you a subtroke of?
			/// </summary>
            public Guid? SubstrokeOf;

			/// <summary>
			/// P1
			/// </summary>
            public string P1;
			
			/// <summary>
			/// P2
			/// </summary>
            public string P2;

			/// <summary>
			/// X-Coordinate of the Shape
			/// </summary>
            public float? X;

			/// <summary>
			/// Y-Coordinate of the Shape
			/// </summary>
            public float? Y;

			/// <summary>
			/// Associated text
			/// </summary>
            public string Text;
			/// <summary>
			/// LeftX
			/// </summary>
            public float? LeftX;
			/// <summary>
			/// TopY
			/// </summary>
            public float? TopY;
			/// <summary>
			/// Control1
			/// </summary>
            public string Control1;
			/// <summary>
			/// Control 2
			/// </summary>
            public string Control2;
			/// <summary>
			/// Starting substroke
			/// </summary>
            public Guid? Start;
			/// <summary>
			/// Ending substroke
			/// </summary>
            public Guid? End;
			/// <summary>
			/// Source (i.e., who made this label?)
			/// </summary>
            public string Source;
			/// <summary>
			/// Pen width setting
			/// </summary>
            public float? PenWidth;
			/// <summary>
			/// Pen height setting
			/// </summary>
            public float? PenHeight;

			/// <summary>
			/// Label Probability of the Shape
			/// </summary>
            public float? Probability;

            /// <summary>
            /// Classification of the shape 
            /// </summary>
            public string Classification;

            /// <summary>
            /// Probability of the classification
            /// </summary>
            public float? ClassificationBelief;

            /// <summary>
            /// The dictionary of connected shapes.
            /// </summary>
            public Dictionary<Shape, string> Connections;

            #endregion

            #region CONSTRUCTORS

			/// <summary>
			/// Create a new XmlShapeAttrs
			/// </summary>
			/// <param name="newGUID">Create a new GUID for this shape?</param>
            public XmlShapeAttrs(bool newGUID)
                :this()
            {
                if (newGUID)
                    Id = Guid.NewGuid();
                else
                    Id = null;
                Name = null;
                Connections = new Dictionary<Shape, string>();
            }

			/// <summary>
			/// Copy constructor
			/// </summary>
			/// <param name="other">The source</param>
            public XmlShapeAttrs(XmlShapeAttrs other)
            {
                Id = other.Id;
                Name = other.Name;
                Time = other.Time;
                Type = other.Type;
                Author = other.Author;
                Color = other.Color;
                Height = other.Height;
                Width = other.Width;
                Area = other.Area;
                LaysInk = other.LaysInk;
                Orientation = other.Orientation;
                PenTip = other.PenTip;
                Raster = other.Raster;
                SubstrokeOf = other.SubstrokeOf;
                P1 = other.P1;
                P2 = other.P2;
                X = other.X;
                Y = other.Y;
                Text = other.Text;
                LeftX = other.LeftX;
                TopY = other.TopY;
                Control1 = other.Control1;
                Control2 = other.Control2;
                Start = other.Start;
                End = other.End;
                Source = other.Source;
                PenWidth = other.PenWidth;
                PenHeight = other.PenHeight;
                Probability = other.Probability;
                Classification = other.Classification;
                ClassificationBelief = other.ClassificationBelief;
                if (other.Connections != null)
                    Connections = other.Connections;
                else
                    Connections = new Dictionary<Shape, string>();
            }

            #endregion

            #region METHODS

            /// <summary>
			/// Get the attribute names.
			/// </summary>
			/// <returns>A string array of attribute names</returns>
			public string[] getAttributeNames()
			{
				return new string[] { "type", "name", "id", "time", "x", "y", "probability", "author", "color", 
										"height", "width", "area", "laysInk", "orientation", "penTip", "penWidth", 
                                        "penHeight", "raster", "substrokeOf", "p1", "p2", "text", "leftx", 
                                        "topy", "control1", "control2", "start", "end", "source", "classification", 
                                        "classificationBelief", "connections" };
			}

			
			/// <summary>
			/// Get the attribute values.
			/// </summary>
			/// <returns>An object array of attribute values</returns>
			public object[] getAttributeValues()
			{
				return new object[] {	Type, Name, Id, Time, X, Y, Probability, Author, Color,
										Height, Width, Area, LaysInk, Orientation, PenTip, PenWidth,
                                        PenHeight, Raster, SubstrokeOf, P1, P2, Text, LeftX, 
                                        TopY, Control1, Control2, Start, End, Source, Classification, ClassificationBelief, Connections };
			}

			
			/// <summary>
			/// Clone method.
			/// </summary>
			/// <returns>A new XmlShapeAttrs struct copying the old one's values</returns>
			public XmlShapeAttrs Clone()
			{
				XmlShapeAttrs temp = new XmlShapeAttrs(false);
				temp.Id = this.Id;
				temp.Name = this.Name;
				temp.Time = this.Time;
				temp.Type = this.Type;
				temp.Author = this.Author;
				temp.Color = this.Color;
				temp.Height = this.Height;
				temp.Width = this.Width;
				temp.Area = this.Area;
				temp.LaysInk = this.LaysInk;
				temp.Orientation = this.Orientation;
				temp.PenTip = this.PenTip;
                temp.PenWidth = this.PenWidth;
                temp.PenHeight = this.PenHeight;
				temp.Raster = this.Raster;
				temp.SubstrokeOf = this.SubstrokeOf;
				temp.P1 = this.P1;
				temp.P2 = this.P2;
				temp.X = this.X;
				temp.Y = this.Y;
                temp.Probability = this.Probability;
				temp.Text = this.Text;
				temp.LeftX = this.LeftX;
				temp.TopY = this.TopY;
				temp.Control1 = this.Control1;
				temp.Control2 = this.Control2;
				temp.Start = this.Start;
				temp.End = this.End;
				temp.Source = this.Source;
                temp.ClassificationBelief = this.ClassificationBelief;
                temp.Classification = this.Classification;
                temp.Connections = this.Connections;

				return temp;
			}

			
			#endregion
		}

		
		/// <summary>
		/// XML attribute struct for the Point data
		/// </summary>
		[Serializable]
		public struct XmlPointAttrs
		{
			#region	INTERNALS
			/// <summary>
			/// X-Coordinate of the Shape
			/// </summary>
            public float? X;
			/// <summary>
			/// Y-Coordinate of the Shape
			/// </summary> 
            public float? Y;

			/// <summary>
			/// Pen Pressure of the Point
			/// </summary>
            public ushort? Pressure;

			/// <summary>
			/// Time the Point was drawn
			/// </summary>
            public ulong? Time;

			/// <summary>
			/// Name of the Point
			/// </summary>
            public string Name;

			/// <summary>
			/// The Point's GUID
			/// </summary>
            public Guid? Id;

			#endregion

			#region METHODS

			/// <summary>
			/// Get the attribute names.
			/// </summary>
			/// <returns>A string array of attribute names</returns>
			public string[] getAttributeNames()
			{
				return new string[] { "x", "y", "pressure", "time", "name", "id" };
			}

						
			/// <summary>
			/// Get the attribute values.
			/// </summary>
			/// <returns>An object array of attribute values</returns>
			public object[] getAttributeValues()
			{
				return new object[] { X, Y, Pressure, Time, Name, Id };
			}

			
			/// <summary>
			/// Clone method.
			/// </summary>
			/// <returns>A new XmlPointAttrs struct copying the old one's values</returns>
			public XmlPointAttrs Clone()
			{
				XmlPointAttrs temp = new XmlPointAttrs();
				temp.X = this.X;
				temp.Y = this.Y;
				temp.Pressure = this.Pressure;
				temp.Time = this.Time;
				temp.Id = this.Id;
				temp.Name = this.Name;
				
				return temp;
			}
			
			#endregion		
		}
	}
}
